import gdown, os, random
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from . import settings, utils, models
from IPython import display
from PIL import Image 

class Dataset():
    def __init__(self, ds_name):
        self.dataset_name = ds_name.lower() # название датасета
        self.sampling = [[],[],[],[]] # выборки
        
    def load_dataset(self):
        print('Загрузка данных')
        print('Это может занять несколько минут...')
        archive_name = settings.param_datasets[self.dataset_name]['archive']
        url = settings.SERVER + '/' +archive_name
        gdown.download(url+'.zip', archive_name, quiet=True) # Скачиваем файл по указанному URL
        gdown.download(url+'_sampling.zip', archive_name+'_sampling', quiet=True) # Скачиваем файл по указанному URL
        utils.unzip_archive(
            from_ = archive_name,
            to_ = os.path.join(self.dataset_name)
            )
        utils.unzip_archive(
            from_ = archive_name+'_sampling',
            to_ = ''
            )
        display.clear_output(wait=True)
        print('Загрузка данных завершена \n')
        print('url:', url)
        print(settings.param_datasets[self.dataset_name]['info'])
    
    def show_samples(self):
        # молочная продукция
        if self.dataset_name == settings.DATASET_NAMES[0]:
            if not os.path.exists(self.dataset_name):
                print(f'Загрузите базу для \'{self.dataset_name}\' для корректной работы.')
                return
            size=(256, 512)
            fig = plt.figure(figsize=(10,10))
            classes = sorted(os.listdir(self.dataset_name))
            for i in range(len(classes)):
                images = sorted(os.listdir(os.path.join(self.dataset_name, classes[i])))
                img_path = os.path.join(self.dataset_name, classes[i], random.choice(images))
                img = Image.open(img_path)  # Открытие обрабатываемого файла
                img = img.resize(size)                
                img = np.array(img)       
                ax = fig.add_subplot(1, 3, i+1, xticks=[], yticks=[])
                plt.imshow(img)
        # автобусы
        elif self.dataset_name == settings.DATASET_NAMES[1]:
            if not os.path.exists(self.dataset_name):
                print(f'Загрузите базу для \'{self.dataset_name}\' для корректной работы.')
                return        
            size=(150, 256)
            entering_idx = [1,170,338,1200,1536,
                      1762,1830,1942,2312,2414,
                      2455,2712,3513,3657,3744,
                      3761,3792,3797,3826,3877,
                      3904,3945,3972,3993,4140]
            getting_off_idx = [184,445,653,1220,1241,
                        1283,1367,1388,1409,1453,
                        1487,1648,1906,2505,2412,
                        2292,1865]
            fig = plt.figure(figsize=(16,10))
            entering_img = sorted(os.listdir(os.path.join(self.dataset_name, 'entering')))
            for i, idx in enumerate(np.random.choice(entering_idx, 5, replace=False)):
                img_path = os.path.join(self.dataset_name, 'entering', entering_img[idx])
                img = Image.open(img_path)  # Открытие обрабатываемого файла
                img = img.resize(size)     
                img = np.array(img)          
                ax = fig.add_subplot(2, 5, i+1, xticks=[], yticks=[])
                ax.set_title('Входящий')
                plt.imshow(img)
            getting_off_img = sorted(os.listdir(os.path.join(self.dataset_name, 'getting_off')))
            for i, idx in enumerate(np.random.choice(getting_off_idx, 5, replace=False)):
                img_path = os.path.join(self.dataset_name, 'getting_off', getting_off_img[idx])
                img = Image.open(img_path)  # Открытие обрабатываемого файла
                img = img.resize(size)
                img = np.array(img)          
                ax = fig.add_subplot(2, 5, i+6, xticks=[], yticks=[])
                ax.set_title('Выходящий')
                plt.imshow(img)
         # квартиры
        elif self.dataset_name == settings.DATASET_NAMES[2]:
            def getRoomsCount(d, maxRoomCount):
                roomsCountStr = d[0] #Получаем строку с числом комнат
                roomsCount = 0
                try:
                    roomsCount = int(roomsCountStr) #Пробуем превратить строку в число
                    if (roomsCount > maxRoomCount): 
                        roomsCount = maxRoomCount #Если число комнат больше максимального, то присваиваем максимальное
                except: #Если не получается превратить строку в число
                    if (roomsCountStr == roomsCountStr): #Проверяем строку на nan (сравнение с самим собой)
                        if ("Ст" in roomsCountStr): #Еcть строка = "Ст", значит это Студия
                            roomsCount = maxRoomCount + 1
                return roomsCount
            df = pd.read_csv('квартиры/moscow.csv', sep=";") #Загружаем данные в data frame
            df = df.iloc[::2,:] #Выбираем нечётные строки, в чётных строках в исходном фрейме пустые строки для комментариев
            data = df.values #Вытаскиваем данные в numpy array
            oneRoomMask = [getRoomsCount(d, 30) == 1 for d in data] #Делаем маску однокомнатных квартир, принцип (getRoomsCount(d, 30) == 1)
            data1 = data[oneRoomMask] #В data1 оставляем только однокомнатные квартиры
            idx = 0
            while idx<5:
                try:
                    i = np.random.randint(0, data1.shape[0])
                    print('* Пример объявления: ')
                    print()
                    print('Количество комнат: ', data1[i][0])
                    print('Площадь квартиры:  ', data1[i][6])
                    print('Метро/ЖД станция:  ', data1[i][1])
                    print('От станции:        ', data1[i][2])
                    print('Дом:               ', data1[i][3])
                    print('Балкон:            ', str(data1[i][4]).replace('nan',''))
                    print('Санузел:           ', str(data1[i][5]).replace('nan',''))
                    s = data1[i][13].replace('/n',' ')
                    s = s.replace('/r',' ')
                    print('Примечание:        ', s)
                    print('-----------')
                    c1 = int(int(data1[i][7])/1e+6)      
                    c2 = int(((int(data1[i][7])/1e+6) - c1) * 1e+3)
                    print('Цена квартиры:     ',  c1, ' млн  ', c2,' тыс рублей', sep='')
                    print('---------------------------------------------------------------------------------------------------------------')
                    print()
                    idx += 1
                except:
                    continue
                
    def create_sampling(self):
        # молочная продукция
        if self.dataset_name == settings.DATASET_NAMES[0] or self.dataset_name == settings.DATASET_NAMES[1]:
            print('Создание наборов данных для обучения модели...')            
            self.sampling[0] = np.load('x_train.npy') # Добавляем в список x_train сформированные данные
            self.sampling[1] = np.load('y_train.npy') # Добавлеям в список y_train значение 0-го класса
            self.sampling[2] = np.load('x_test.npy') # Добавляем в список x_test сформированные данные
            self.sampling[3] = np.load('y_test.npy')# Добавлеям в список y_test значение 0-го класса
            display.clear_output(wait=True)
            print('Размер сформированного массива обучающая_выборка:'.ljust(57), self.sampling[0].shape)
            print('Размер сформированного массива метки_обучающей_выборки:'.ljust(57), self.sampling[1].shape)
            print('Размер сформированного массива тестовая_выборка:'.ljust(57), self.sampling[2].shape)
            print('Размер сформированного массива метки_тестовой_выборки:'.ljust(57), self.sampling[3].shape)
        # квартиры
        if self.dataset_name == settings.DATASET_NAMES[2]:
            print('Создание наборов данных для обучения модели...')            
            self.sampling[0] = (np.load('flats_x_train_1.npy'), np.load('flats_x_train_2.npy')) # Добавляем в список x_train сформированные данные
            self.sampling[1] = np.load('flats_y_train.npy') # Добавлеям в список y_train значение 0-го класса
            self.sampling[2] = (np.load('flats_x_test_1.npy'), np.load('flats_x_test_2.npy')) # Добавляем в список x_test сформированные данные
            self.sampling[3] = np.load('flats_y_test.npy')# Добавлеям в список y_test значение 0-го класса
            display.clear_output(wait=True)
            print('Размер сформированного массива обучающая_выборка (числовые данные):'.ljust(77), self.sampling[0][0].shape)
            print('Размер сформированного массива обучающая_выборка (текстовые данные):'.ljust(77), self.sampling[0][1].shape)
            print('Размер сформированного массива метки_обучающей_выборки:'.ljust(77), self.sampling[1].shape)
            print('Размер сформированного массива тестовая_выборка(числовые данные):'.ljust(77), self.sampling[2][0].shape)
            print('Размер сформированного массива тестовая_выборка(текстовые данные):'.ljust(77), self.sampling[2][1].shape)
            print('Размер сформированного массива метки_тестовой_выборки:'.ljust(77), self.sampling[3].shape)

    def create_model(self, layers):
        self.terraModel = models.TerraModel(layers, settings.param_datasets[self.dataset_name])

    def train_model(self, epochs, repeats):
        if settings.param_datasets[self.dataset_name]['task'] == 'классификация изображений':
            self.terraModel.train_model(epochs, repeats, self.sampling)
        elif settings.param_datasets[self.dataset_name]['task'] == 'регрессия квартир':
            self.terraModel.train_fmodel(epochs, self.sampling)
        
    def test_model(self, data):
        self.terraModel.test_model(data, settings.param_datasets[self.dataset_name], self.sampling[2:])


from tensorflow.keras.models import Sequential, Model
from tensorflow.keras.layers import Input, Flatten, Conv2D, Dense, concatenate, Activation # Импортируем стандартные слои keras
from tensorflow.keras.callbacks import ModelCheckpoint, LambdaCallback, ReduceLROnPlateau
from tensorflow.keras.optimizers import Nadam
import time, random, gc, pickle
import numpy as np
import pandas as pd
from IPython import display
from termcolor import colored
import matplotlib.pyplot as plt
from . import utils

class TerraModel():    
    def __init__(self, layers, param):
        self.model = None
        self.input_size = param['sampling_size']
        self.layers = layers
        if isinstance(layers, tuple):            
            self.create_fmodel(param['task'])
        else:            
            self.create_model(param['task'])
        
    def create_model(self, task, rep = False):
        task = task.lower()
        self.task = task
        if task == 'классификация изображений':
            params = {'loss':'sparse_categorical_crossentropy',
                      'opt':'adam',
                       'metrics':['accuracy']}
        self.create_image_classification_model(
            self.layers + '-softmax',
            params,
            rep
        )
        
    def create_fmodel(self, task, rep = False):
        task = task.lower()
        self.task = task
        input1 = Input(self.input_size[0],)
        input2 = Input(self.input_size[1],)
                  
        layers = self.layers[0].split()
        x1 = self.create_layer(layers[0]) (input1)
        layers[-1]+='-linear'
        for i in range(1, len(layers)):
            layer = self.create_layer(layers[i])
            assert layer!=0, 'Невозможно добавить указанный слой: '+layer
            x1 = self.create_layer(layers[i]) (x1)

        layers = self.layers[1].split()
        x2 = self.create_layer(layers[0]) (input2)
        layers[-1]+='-linear'
        for i in range(1, len(layers)):
            layer = self.create_layer(layers[i])
            assert layer!=0, 'Невозможно добавить указанный слой: '+layer
            x2 = self.create_layer(layers[i]) (x2)
           
        x = concatenate([x1, x2])        
        layers = self.layers[2].split()        
        layers[-1]+='-linear'
        x3 = self.create_layer(layers[0]) (x)
        for i in range(1, len(layers)):
            layer = self.create_layer(layers[i])
            assert layer!='0', 'Невозможно добавить указанный слой: '+layer
            x3 = self.create_layer(layers[i]) (x3)        
        self.model = Model([input1, input2], x3)
        self.model.compile(loss="mae", optimizer=Nadam(learning_rate=1e-3), metrics=["mae"])        
        print('Создана модель нейронной сети!')
        
    def create_image_classification_model(self, layers, params, rep):
        self.model = Sequential()        
        layers = layers.split()
        layer = self.create_layer(layers[0], входной_размер=self.input_size)
        self.model.add(layer)   
        for i in range(1, len(layers)):
            layer = self.create_layer(layers[i])
            self.model.add(layer)
        if not rep:
            print('Создана модель нейронной сети!')
        if 'metrics' in params:
            self.model.compile(loss=params['loss'], optimizer = params['opt'], metrics=params['metrics'])
        else:
            self.model.compile(loss=params['loss'], optimizer = params['opt'])
    
    def create_layer(self, layer, **kwargs):
        args={}
        if 'входной_размер' in kwargs:
            args['input_shape'] = kwargs['входной_размер']  
        параметры = [layer]
        act = 'relu'
        if '-' in layer:
            параметры = layer.split('-')
        if параметры[0].upper() == 'ПОЛНОСВЯЗНЫЙ':    
            if len(параметры)>2:
                act = параметры[2]
            return Dense(int(параметры[1]), activation=act, **args)
        elif параметры[0].upper() == 'СВЕРТОЧНЫЙ2D':
            if len(параметры)<5:
                act = 'relu'
                pad='same'
            else:
                act = параметры[4]
                pad = параметры[3]
            if any(i in '()' for i in параметры[2]):
                return Conv2D(int(параметры[1]), (int(параметры[2][1]),int(параметры[2][3])), padding=pad,activation=act, **args)
            else:
                return Conv2D(int(параметры[1]), int(параметры[2]), padding=pad,activation=act, **args)
        elif параметры[0].upper() == 'ВЫРАВНИВАЮЩИЙ':
            if 'input_shape' in args:
                return Flatten(input_shape=args['input_shape'])
            else:
                return Flatten()   
        else:
            assert False, f'Невозможно добавить указанный слой: \'{параметры[0]}\'.\
                Возможно вы имели ввиду \'{check_for_errors.check_word(параметры[0], "layer")}\'.'
                
    def train_model(self, epochs, repeats, sampling=[]):
        x_train = sampling[0]
        x_test = sampling[2]
        y_train = sampling[1]
        y_test = sampling[3]
        batch_size = 64
        global result, best_result, idx_best, best_result_on_train_arr
        cur_time = time.time()
        
        if repeats > 0:
            cur_time = time.time()
            global stage_result, stage_best_result
            stage_result = ''
            stage_best_result = []
            best_result_on_train_arr = []
            stage_history = []
            for start in range(repeats):
                del(self.model)
                gc.collect()
                self.create_model(self.task, rep = True)
                result = ''
                idx_best = 0
                best_result = 0
                best_result_on_train = 0
                filepath="model.h5"

                model_checkpoint_callback = ModelCheckpoint(
                    filepath=filepath,
                    save_weights_only=True,
                    monitor='val_loss',
                    mode='min',
                    save_best_only=True)
                cur_time = time.time()
                
                def on_epoch_end(epoch, log):
                    global cur_time, result, best_result, idx_best, stage_result, текущая_эпоха, stage_best_result, best_result_on_train, best_result_on_train_arr
                    текущая_эпоха = epoch
                    k = list(log.keys()) 
                    p1 = 'Запуск №' + str(start+1)
                    p2 = p1 + ' '* (12 - len(p1)) + 'Время обучения: ' + str(round((time.time()-cur_time) * epochs,1)) +'c'
                    p3 = p2 + ' '* (34 - len(p2)) + 'Точность на обучающей выборке: ' + str(round(log[k[1]]*100,1))+'%'
                    if len(k)>2:
                        p4 = p3 + ' '* (77 - len(p3)) + 'Точность на проверочной выборке: ' + str(round(log[k[3]]*100,1))+'%'
                        result += p4 + '\n'
                        if log[k[3]]*100 >= best_result:
                            best_result = log[k[3]]*100
                            best_result_on_train = log[k[1]]*100
                            idx_best = epoch
                    else:
                        result += p3 + '\n'
                        if log[k[1]]*100 >= best_result:
                            best_result = log[k[1]]*100
                            best_result_on_train = log[k[1]]*100
                            idx_best = epoch 
                    if (текущая_эпоха+1) == epochs:
                        result = result.split('\n')
                        stage_best_result.append(best_result)
                        best_result_on_train_arr.append(best_result_on_train)
                        for i in range(len(result)):
                            if i == idx_best:
                                best_in_epoc = result[i]
                                stage_result += best_in_epoc + '\n'
                        print(best_in_epoc)
                    cur_time = time.time()

                def on_epoch_begin(epoch, log):
                    global cur_time
                    cur_time = time.time()
                    
                myCB = LambdaCallback(on_epoch_end = on_epoch_end, on_epoch_begin=on_epoch_begin)              
                history = self.model.fit(x_train, y_train, batch_size=batch_size, epochs=epochs, validation_data = (x_test, y_test), callbacks=[model_checkpoint_callback, myCB], verbose = 0)
                stage_history.append(history)
                self.model.load_weights('model.h5')
                self.model.save('model_s.h5')

            display.clear_output(wait=True)
            stage_result = stage_result.split('\n')
            stage_best_result = np.asarray(stage_best_result)
            best_result_on_train_arr = np.asarray(best_result_on_train_arr)
            idx_best_stage = np.argmax(stage_best_result)

            for i in range(len(stage_result)):
                s = stage_result[i]
                if i == idx_best_stage:
                    s = colored(stage_result[i], color='white', on_color='on_green')
                print(s)
            print('Средняя точность на обучающей выборке:   ' , str(round(np.mean(best_result_on_train_arr), 1)) +  '%', '\n' 
                  'Средняя точность на проверочной выборке: ' , str(round(np.mean(stage_best_result), 1)) + '%')

            keys = list(stage_history[idx_best_stage].history.keys())
            data={
                "Обучающая выборка": best_result_on_train_arr,
                "Проверочная выборка": stage_best_result
                }
            df=pd.DataFrame(data)
            df.plot(kind="bar",stacked=False,figsize=(10,8))
            plt.legend(loc="lower left",bbox_to_anchor=(0.8,1.0))
            plt.hlines(round(np.mean(stage_best_result)), -1, len(stage_history), color='green')
            plt.title('График точности обучения, на лучшем запуске') # Выводим название графика         
            plt.show()


        else:
            result = ''
            idx_best = 0
            best_result = 0            
            model_checkpoint_callback = ModelCheckpoint(
                filepath="model.h5",
                save_weights_only=True,
                monitor='val_loss',
                mode='min',
                save_best_only=True)              
            cur_time = time.time()
            def on_epoch_end(epoch, log):
                k = list(log.keys())
                global cur_time, result, idx_best, best_result
                p = f'Эпоха №{epoch+1}'.ljust(12) \
                    + f'Время обучения: {round(time.time()-cur_time,1)}c'.ljust(24) \
                    + f'Точность на обучающей выборке: {round(log[k[1]]*100,1)}%'.ljust(44) \
                    + f'Точность на проверочной выборке: {round(log[k[3]]*100,1)}%'
                result += p + '\n'
                if log[k[3]]*100 >= best_result:
                    best_result = log[k[3]]*100
                    idx_best = epoch
                print(p)
                cur_time = time.time()

            def on_epoch_begin(epoch, log):
                global cur_time
                cur_time = time.time()

            myCB = LambdaCallback(on_epoch_end = on_epoch_end, on_epoch_begin=on_epoch_begin)            
            history = self.model.fit(x_train, y_train, batch_size=batch_size, epochs=epochs, validation_data = (x_test, y_test), callbacks=[model_checkpoint_callback, myCB], verbose = 0)
            
            self.model.load_weights('model.h5')
            self.model.save('model_s.h5')
            display.clear_output(wait=True)
            result = result.split('\n')
            for i in range(len(result)):
                s = result[i]
                if i == idx_best:
                    s = colored(result[i], color='white', on_color='on_green')
                print(s)
            keys = list(history.history.keys())
            plt.plot(history.history[keys[1]], label ='Обучающая выборка') # Визуализируем график точности на обучающей выборке
            if len(keys)>2:
                plt.plot(history.history['val_'+keys[1]], label ='Проверочная выборка') # Визуализируем график точности на проверочной выборке
            plt.legend() # Выводим подписи на графике
            plt.title('График точности обучения') # Выводим название графика
            plt.show()
            return history
    
    def train_fmodel(self, epochs, sampling=[]):
        cur_time = time.time()
        global loss, val_loss, history, result, best_result, idx_best
        result = ''
        idx_best = 0
        best_result = 50000000
        filepath="model.h5"
        batch_size = 256
        model_checkpoint_callback = ModelCheckpoint(
            filepath=filepath,
            save_weights_only=True,
            monitor='val_loss',
            mode='min',
            save_best_only=True)

        def on_epoch_end(epoch, log):
            global cur_time, loss, val_loss, result, best_result, idx_best, count_len
            yScaler = pickle.load(open('yscaler.pkl','rb'))
            pred = self.model.predict(sampling[2]) #Полуаем выход сети на проверочно выборке
            predUnscaled = yScaler.inverse_transform(pred).flatten() #Делаем обратное нормирование выхода к изначальным величинам цен квартир
            yTrainUnscaled = yScaler.inverse_transform(sampling[3]).flatten() #Делаем такое же обратное нормирование yTrain к базовым ценам
            delta = predUnscaled - yTrainUnscaled #Считаем разность предсказания и правильных цен
            absDelta = abs(delta) #Берём модуль отклонения

            pred2 = self.model.predict(sampling[0]) #Полуаем выход сети на проверочно выборке
            predUnscaled2 = yScaler.inverse_transform(pred2).flatten() #Делаем обратное нормирование выхода к изначальным величинам цен квартир
            yTrainUnscaled2 = yScaler.inverse_transform(sampling[1]).flatten() #Делаем такое же обратное нормирование yTrain к базовым ценам
            delta2 = predUnscaled2 - yTrainUnscaled2 #Считаем разность предсказания и правильных цен
            absDelta2 = abs(delta2) #Берём модуль отклонения
            loss.append(sum(absDelta2) / (1e+6 * len(absDelta2)))
            val_loss.append(sum(absDelta) / (1e+6 * len(absDelta)))
            
            p1 = 'Эпоха №' + str(epoch+1)
            p2 = p1 + ' '* (10 - len(p1)) + 'Время обучения: ' + str(round(time.time()-cur_time,2)) +'c'
            p3 = p2 + ' '* (33 - len(p2)) + 'Ошибка на обучающей выборке: ' + str(round(sum(absDelta2) / (1e+6 * len(absDelta2)), 3))+'млн'
            p4 = p3 + ' '* (77 - len(p3)) + 'Ошибка на проверочной выборке: ' + str(round(sum(absDelta) / (1e+6 * len(absDelta)), 3))+'млн'
            result += p4 + '\n' 
            
            a = round(sum(absDelta) / (1e+6 * len(absDelta)), 3)
            if a < best_result:
                best_result = a
                idx_best = epoch
            print(p4)           

        def on_train_begin(log):
            global cur_time, loss, val_loss
            loss=[]
            val_loss = []

        def on_epoch_begin(epoch, log):
            global cur_time
            cur_time = time.time()

        reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.2,
                                      patience=5, min_lr=1e-6)

        myCB = LambdaCallback(on_train_begin=on_train_begin, on_epoch_end = on_epoch_end, on_epoch_begin=on_epoch_begin)
        myCB23 = LambdaCallback(on_epoch_end = on_epoch_end, on_epoch_begin=on_epoch_begin)
        self.model.fit(sampling[0], sampling[1], batch_size=batch_size, epochs=epochs, validation_data = (sampling[2], sampling[3]), callbacks=[model_checkpoint_callback, myCB, reduce_lr], verbose = 0)
        self.model.load_weights('model.h5')
        self.model.save('model_s.h5')

        display.clear_output(wait=True)
        result = result.split('\n')
        self.res = result
        self.l= loss
        self.vl = val_loss
        idx = 0
        for i in range(len(result)):
            s = result[i]
            if i == idx_best:
                s = colored(result[i], color='white', on_color='on_green')
            print(s)
        plt.figure(figsize=(12, 6)) # Создаем полотно для визуализации  
        plt.plot(loss, label ='Обучающая выборка') # Визуализируем график ошибки на обучающей выборке
        plt.plot(val_loss, label ='Проверочная выборка') # Визуализируем график ошибки на проверочной выборке
        plt.legend() # Выводим подписи на графике
        plt.title('График ошибки обучения') # Выводим название графика
        plt.show()
  
    def test_model(self, data, param, test_sampling):
        if param['task']=='классификация изображений':
            self.test_model_image_classification(param, test_sampling)
        if param['task']=='регрессия квартир':
            self.test_model_regression(data)
        
    def test_model_image_classification(self, param, test_sampling):        
        for i in range(5):
            number = np.random.randint(test_sampling[0].shape[0]) # Задаем индекс изображения в тестовом наборе
            sample = test_sampling[0][number]            
            print('Тестовое изображение:')
            plt.imshow(sample, cmap='gray') # Выводим изображение из тестового набора с заданным индексом
            plt.axis('off') # Отключаем оси
            plt.show()
            sample = sample.reshape((1 + self.model.input.shape[1:]))
            pred = self.model.predict(sample)[0] # Распознаем изображение с помощью обученной модели
            print()
            print('Результат предсказания модели:')
            dicts = {}
            classes = param['classes']
            for i in range(len(classes)):
                dicts[classes[i]] = round(100*pred[i],2)
                print(f'   класс {classes[i]}:'.ljust(40), round(100*pred[i],1),'%',sep='')
            print('---------------------------')
            answer = str(classes[test_sampling[1][number]])
            if len(test_sampling[1])>0:
                if utils.keywithmaxval(dicts) == answer:
                    print('Правильный ответ: ', utils.out_green(answer))
                    print('---------------------------')
                    print()
                    print()
                elif utils.keywithmaxval(dicts) != classes[test_sampling[1][number]]:
                    print('Правильный ответ: ', utils.out_red(answer))
                    print('---------------------------')
                    print()
                    print()  
        
    def test_model_regression(self, данные):
        df = pd.DataFrame(
            np.array([[1, данные[0], str(данные[1])+данные[2][0], str(данные[3])+'/'+str(данные[4])+' П', данные[5],  данные[6], str(данные[7])+'/12/12', 0, 0,0,0,0,0,данные[8] ]]),
            columns=['Комнат', 'Метро / ЖД станции', 'От станции', 'Дом', 'Балкон', 'Санузел', 'Площадь', 'Цена, руб.', 'ГРМ', 'Бонус агенту', 'Дата', 'Кол-во дней в экспозиции', 'Источник', 'Примечание'],
        )
        data = df.values #Вытаскиваем данные в numpy array
        oneRoomMask = [utils.getRoomsCount(d, 30) == 1 for d in data] #Делаем маску однокомнатных квартир, принцип (getRoomsCount(d, 30) == 1)
        data1 = data[oneRoomMask] #В data1 оставляем только однокомнатные квартиры
        xTrain = utils.getXTrain(data1, df)
        yTrain = utils.getYTrain(data1)
        xTrainC, allTextComments = utils.getXTrainComments(data1) #Создаём обучающую выборку по текстам и большо текст для словаря
        allWords = utils.text2Words(allTextComments) #Собираем полный текст в слова
        allWords = allWords[::10] #Берём 10% слов (иначе словарь слишком долго формируется)
        vocabulary = utils.createVocabulary(allWords) #Создаём словарь
        xTrainC01 = utils.changeSetToIndexes(xTrainC, vocabulary, 2000) #Преобразеум xTrain в bag of words
        #Нормируем размер квартиры в xTrain  
        xTrainScaled = xTrain.copy()
        xScaler = pickle.load(open('xscaler.pkl','rb'))
        yScaler = pickle.load(open('yscaler.pkl','rb'))
        xTrainScaled[:,-1] = xScaler.transform(xTrain[:,-1].reshape(-1, 1)).flatten() #Нормируем данные нормировщиком
        
        pred = self.model.predict([xTrainScaled, xTrainC01]) #Полуаем выход сети на проверочно выборке
        predUnscaled = yScaler.inverse_transform(pred).flatten() #Делаем обратное нормирование выхода к изначальным величинам цен квартир
        print('Цена квартиры: ' + str(int(predUnscaled[0]//1000000))+'млн  '+str(int(predUnscaled[0]%1000000//1000))+'тыс рублей')


                


SERVER = 'https://storage.googleapis.com/terra_ai/DataSets'

DATASET_NAMES = ('молочная_продукция', 'автобусы', 'квартиры')
param_datasets = {    
    DATASET_NAMES[0] : 
        {'archive' : 'milk',
        'info' : 'Вы скачали базу с изображениями бутылок молока',
        'task' : 'классификация изображений',
        'sampling_size' : (96, 53, 3),
        'classes' : ['Parmalat', 'Кубанский_молочник', 'Кубанская_буренка']
        },
    DATASET_NAMES[1] : 
        {'archive' : 'bus',
        'info' : 'Вы скачали базу с изображениями входящих и выходящих пассажиров в автобус',
        'task' : 'классификация изображений',
        'sampling_size' : (128, 64, 3),
        'classes' : ['Входящий', 'Выходящий']
        },
    DATASET_NAMES[2] : 
        {'archive' : 'flats',
        'info' : 'Загружена база квартир',
        'task' : 'регрессия квартир',
        'sampling_size' : (109, 2000),
        },
    }


import zipfile
from tensorflow.keras.utils import to_categorical #Используем для to_categoricall
import numpy as np
import gc

def unzip_archive(from_='', to_=''):    
    with zipfile.ZipFile(from_, 'r') as zip_ref:
        zip_ref.extractall(to_)
        
def out_red(text):
    return "\033[4m\033[31m\033[31m{}\033[0m".format(text)
    
def out_green(text):
    return "\033[4m\033[32m\033[32m{}\033[0m".format(text)
    
def keywithmaxval(d):
    v = list(d.values())
    k = list(d.keys())
    return str(k[v.index(max(v))])

def getRoomsCount(d, maxRoomCount):
  roomsCountStr = d[0] #Получаем строку с числом комнат
  roomsCount = 0
  try:
    roomsCount = int(roomsCountStr) #Пробуем превратить строку в число
    if (roomsCount > maxRoomCount): 
      roomsCount = maxRoomCount #Если число комнат больше максимального, то присваиваем максимальное
  except: #Если не получается превратить строку в число
    if (roomsCountStr == roomsCountStr): #Проверяем строку на nan (сравнение с самим собой)
      if ("Ст" in roomsCountStr): #Еcть строка = "Ст", значит это Студия
        roomsCount = maxRoomCount + 1
  return roomsCount

#Генерируем обучающаюу выборку - xTrain
def getXTrain(data,df):  
  allMertroNames = list(df["Метро / ЖД станции"].unique())  
  xTrain = [getAllParameters(d, allMertroNames) for d in data]
  xTrain = np.array(xTrain)  
  return xTrain

#Генерируем обучающую выборку - yTrain
def getYTrain(data):  
  costList = [getCost(d) for d in data] 
  yTrain = np.array(costList)  
  return yTrain

def getXTrainComments(data):
  xTrainComments = [] #Тут будет обучающся выборка
  allTextComments = "" #Тут будуте все тексты вместе для словаря
  for d in data:
    currText = getComment(d) #Вытаскиваем примечание к квартире
    try: 
      if (currText == currText): #Проверяем на nan
        allTextComments += currText + " " #Добавляем текст в общий текст для словаря
    except:
      currText = "Нет комментария" #Если не получается, то делаем стандартный текст "Нет комментария"
    xTrainComments.append(currText) #Добавляем примечание новой строкой в обучающую выборку
  xTrainComments = np.array(xTrainComments)
  return (xTrainComments, allTextComments)

def text2Words(text):
  text = text.replace(".", "")# удаляем лишние символы
  text = text.replace("—", "")
  text = text.replace(",", "")
  text = text.replace("!", "")
  text = text.replace("?", "")
  text = text.replace("…", "")
  text = text.lower() #Переводим в нижний регистр
  words = [] #Тут будут все слов
  currWord = "" #Тут будет накапливаться текущее слово, между двумя пробелами
  for symbol in text:
    if (symbol != "\ufeff"): #Игнорируем системынй символ в начале строки
      if (symbol != " "): #Если символ не пробел
        currWord += symbol #То добавляем вимвол в текущее слово
      else: #Если символ пробел
        if (currWord != ""): 
          words.append(currWord) #Добавляем тккущее слово в список слов
          currWord = "" #И обнуляем текущее слово
  if (currWord != ""):
        words.append(currWord)
  return words

def createVocabulary(allWords):
  wCount = dict.fromkeys(allWords, 0)
  for word in allWords:
    wCount[word] += 1 #И увеличиаем количество появлений текущего слова на 1
  wordsList = list(wCount.items())
  wordsList.sort(key = lambda i:i[1], reverse=1)
  sortedWords = [] #Тут будет лист всех отсортированных слов
  for word in wordsList:
    sortedWords.append(word[0]) #Докидываем слово в лист отсортированных слов
  wordIndexes = dict.fromkeys(allWords, 0)
  for word in wordIndexes.keys():
    wordIndexes[word] = sortedWords.index(word)+1 #Ставим индекс = индекс слова в отсортированном листе слов + 1
  return wordIndexes

def changeSetToIndexes(xTrainComments, vocabulary, maxWordsCount):
  xTrainCommentsIndexes = [] #Тут будет итоговый xTrain примечаний в виде индексов
  for text in xTrainComments:
    currWords = text2Words(text) #Разбиваем текст на слова
    currIndexes = words2Indexes(currWords, vocabulary, maxWordsCount) #Превращаем в лист индексов
    currIndexes = np.array(currIndexes)
    xTrainCommentsIndexes.append(currIndexes) #Добавляем в xTrain
  xTrainCommentsIndexes = np.array(xTrainCommentsIndexes)
  xTrainCommentsIndexes = changeSetTo01(xTrainCommentsIndexes, maxWordsCount) #Превращаем в формат bag of words
  return xTrainCommentsIndexes

def getAllParameters(d, allMetroNames):
  gc.collect()
  roomsCountType = getRoomsCountCategory(d, 30)
  metro = getMetro(d, allMetroNames)
  metroType = getMetroType(d)
  metroDistance = getMetroDistance(d)
  floor, floors, isLastFloor, houseType = getHouseTypeAndFloor(d)
  balcony = getBalcony(d)
  wc = getWC(d)
  area = getArea(d)
  out = list(roomsCountType)
  out.append(metro)
  out.extend(metroType)
  out.extend(metroDistance)
  out.extend(floor)
  out.extend(floors)
  out.append(isLastFloor)
  out.extend(houseType)
  out.extend(balcony)
  out.extend(wc)
  out.append(area)
  return out

def getRoomsCountCategory(d, maxRoomCount):
  roomsCount = getRoomsCount(d, maxRoomCount) #Получаем число комнат
  roomsCount = to_categorical(roomsCount, maxRoomCount+2) #Превращаем в категорию
  return roomsCount

def getMetro(d, allMetroNames):
  metroStr = d[1] #Получаем строку метро
  metro = 0
  if (metroStr in allMetroNames): #Если находим метро во всех названиях
    metro = allMetroNames.index(metroStr)+1 #Присваиваем индекс
  return metro

def getMetroType(d):
  metroTypeStr = d[1] #Получаем строку метро
  metroTypeClasses = 5 #Число классов метро
  metroType = metroTypeClasses - 1 #Изначально считаем последний класс
  metroNamesInsideCircle = ["Площадь Революции", "Арбатская", "Смоленская", "Красные Ворота", "Чистые пруды", "Лубянка", "Охотный Ряд", "Библиотека имени Ленина", "Кропоткинская", "Сухаревская", "Тургеневская", "Китай-город", "Третьяковская", "Трубная", "Сретенский бульвар", "Цветной бульвар", "Чеховская", "Боровицкая", "Полянка", "Маяковская", "Тверская", "Театральная", "Новокузнецкая", "Пушкинская", "Кузнецкий Мост", "Китай-город", "Александровский сад"]
  metroNamesCircle = ["Киевская", "Парк Культуры", "Октябрьская", "Добрынинская", "Павелецкая", "Таганская", "Курская", "Комсомольская", "Проспект Мира", "Новослободская", "Белорусская", "Краснопресненская"]
  metroNames13FromCircle = ["Бауманская", "Электрозаводская", "Семёновская", "Площадь Ильича", "Авиамоторная", "Шоссе Энтузиастов", "Римская", "Крестьянская Застава", "Дубровка", "Пролетарская", "Волгоградский проспект", "Текстильщики", "Автозаводская", "Технопарк", "Коломенская", "Тульская", "Нагатинская", "Нагорная", "Шаболовская", "Ленинский проспект", "Академическая", "Фрунзенская", "Спортивная", "Воробьёвы горы", "Студенческая", "Кутузовская", "Фили", "Парк Победы", "Выставочная", "Международная", "Улица 1905 года", "Беговая", "Полежаевская", "Динамо", "Аэропорт", "Сокол", "Деловой центр", "Шелепиха", "Хорошёвская", "ЦСКА", "Петровский парк", "Савёловская", "Дмитровская", "Тимирязевская", "Достоевская", "Марьина Роща", "Бутырская", "Фонвизинская", "Рижская", "Алексеевская", "ВДНХ", "Красносельская", "Сокольники", "Преображенская площадь"]
  metroNames48FromCircle = ["Партизанская", "Измайловская", "Первомайская", "Щёлковская", "Новокосино", "Новогиреево", "Перово", "Кузьминки", "Рязанский проспект", "Выхино", "Лермонтовский проспект", "Жулебино", "Партизанская", "Измайловская", "Первомайская", "Щёлковская", "Новокосино", "Новогиреево", "Перово", "Кузьминки", "Рязанский проспект", "Выхино", "Лермонтовский проспект", "Жулебино", "Улица Дмитриевского", "Кожуховская", "Печатники", "Волжская", "Люблино", "Братиславская", "Коломенская", "Каширская", "Кантемировская", "Царицыно", "Орехово", "Севастопольская", "Чертановская", "Южная", "Пражская", "Варшавская", "Профсоюзная", "Новые Черёмушки", "Калужская", "Беляево", "Коньково", "Университет", "Багратионовская", "Филёвский парк", "Пионерская", "Кунцевская", "Молодёжная", "Октябрьское Поле", "Щукинская", "Спартак", "Тушинская", "Сходненская", "Войковская", "Водный стадион", "Речной вокзал", "Беломорская", "Ховрино", "Петровско-Разумовская", "Владыкино", "Отрадное", "Бибирево", "Алтуфьево", "Фонвизинская", "Окружная", "Верхние Лихоборы", "Селигерская", "ВДНХ", "Ботанический сад", "Свиблово", "Бабушкинская", "Медведково", "Преображенская площадь", "Черкизовская", "Бульвар Рокоссовского"]
  if (metroTypeStr in metroNamesInsideCircle):
    metroType = 0
  if (metroTypeStr in metroNamesCircle):
    metroType = 1
  if (metroTypeStr in metroNames13FromCircle):
    metroType = 2
  if (metroTypeStr in metroNames48FromCircle):
    metroType = 3
  metroType = to_categorical(metroType, metroTypeClasses)
  return metroType

#Вычисляем растояние до метро
def getMetroDistance(d):
  metroDistanceStr = d[2] #Получаем строку
  metroDistance = 0 #Расстояние до метро
  metroDistanceType = 0 #Тип расстояния - пешком или на транспорте
  if (metroDistanceStr == metroDistanceStr):
    if (len(metroDistanceStr) > 0):
      if (metroDistanceStr[-1] == "п"):
        metroDistanceType = 1 #Пешком
      elif (metroDistanceStr[-1] == "т"):
        metroDistanceType = 2 #На транспорте
      metroDistanceStr = metroDistanceStr[:-1]
      try:
        metroDistance = int(metroDistanceStr)
        if (metroDistance < 3):
          metroDistance = 1
        elif (metroDistance < 6):
          metroDistance = 2
        elif (metroDistance < 10):
          metroDistance = 3
        elif (metroDistance < 15):
          metroDistance = 4
        elif (metroDistance < 20):
          metroDistance = 5
        else:
          metroDistance = 6
      except: #Если в строке не число, то категория 0
        metroDistance = 0
  metroDistanceClasses = 7
  if (metroDistanceType == 2):
    metroDistance += metroDistanceClasses #Для типа "Транспортом" добавляем 7
  if (metroDistanceType == 0):
    metroDistance += 2*metroDistanceClasses #Для неопознанного типа добавляем 14
    
  #Превращаем в категории
  metroDistance = to_categorical(metroDistance, 3*metroDistanceClasses)
  return metroDistance

def getHouseTypeAndFloor(d):
  try:
    houseStr = d[3] #Получаем строку типа дома и этажей
  except:
    houseStr = ""
  houseType = 0 #Тип дома
  floor = 0 #Этаж квартиры
  floors = 0 #Этажность дома
  isLastFloor = 0 #Индикатор последнего этажа
  if (houseStr == houseStr):
    if (len(houseStr) > 1):
      try:
        slashIndex = houseStr.index("/") #Ищем разделитель /
      except:
        print(houseStr)
      try:
        spaceIndex = houseStr.index(" ") #Ищем разделитель " "
      except:
        print(houseStr)
      floorStr = houseStr[:slashIndex] #Строка этажа
      floorsStr = houseStr[slashIndex+1:spaceIndex] #Строка этажнгости дома
      houseTypeStr = houseStr[spaceIndex+1:] #Строка типа дома
      try:
        floor = int(floorStr) #Превращаем строку в число
        floorSave = floor
        if (floorSave < 5):
          floor = 2
        if (floorSave < 10):
          floor = 3
        if (floorSave < 20):
          floor = 4
        if (floorSave >= 20):
          floor = 5
        if (floorSave == 1): #Первый этаж выделяем в отдельную категорию
          floor = 1 
        if (floor == floors): #Если этаж последний, включаем индикатор последнего этажа
          isLastFloor = 1 
      except:
        floor = 0 #Если строка не парсится в число, то категория этажа = 0 (отдельная)
      try:
        floors = int(floorsStr) #Превращаем строку в число
        floorsSave = floors
        if (floorsSave < 5):
          floors = 1
        if (floorsSave < 10):
          floors = 2
        if (floorsSave < 20):
          floors = 3
        if (floorsSave >= 20):
          floors = 4
      except:
        floors = 0 #Если строка не парсится в число, то категория этажности = 0 (отдельная)
      if (len(houseTypeStr) > 0):
        if ("М" in houseTypeStr): 
          houseType = 1
        if ("К" in houseTypeStr): 
          houseType = 2
        if ("П" in houseTypeStr): 
          houseType = 3
        if ("Б" in houseTypeStr): 
          houseType = 4
        if ("?" in houseTypeStr): 
          houseType = 5
        if ("-" in houseTypeStr): 
          houseType = 6
    floor = to_categorical(floor, 6)
    floors = to_categorical(floors, 5)
    houseType = to_categorical(houseType, 7)
  return floor, floors, isLastFloor, houseType

#Вычисляем тип балкона
def getBalcony(d):
  balconyStr = d[4] #Полуаем строку
  balconyVariants = ['Л', 'Б', '2Б', '-', '2Б2Л', 'БЛ', '3Б', '2Л', 'Эрк', 'Б2Л', 'ЭркЛ', '3Л', '4Л', '*Л', '*Б']
  if (balconyStr == balconyStr):
    balcony = balconyVariants.index(balconyStr)+1 #Находим индекс строки балкона во всех строках
  else:
    balcony = 0 #Индекс 0 выделяем на строку nan
  balcony = to_categorical(balcony, 16)
  return balcony

def getWC(d):
  wcStr = d[5] #Получаем строку
  wcVariants = ['2', 'Р', 'С', '-', '2С', '+', '4Р', '2Р', '3С', '4С', '4', '3', '3Р']
  if (wcStr == wcStr):
    wc = wcVariants.index(wcStr)+1 #Находим индекс строки санузла во всех строках
  else:
    wc = 0 #Индекс 0 выделяем на строку nan
  wc = to_categorical(wc, 14)
  return wc

#Определяем площадь
def getArea(d):
  areaStr = d[6] #Поулачем строку площади
  if ("/" in areaStr):
    slashIndex = areaStr.index("/") #Находим разделитель /
    try:
      area = float(areaStr[:slashIndex]) #Берём число до разделителя и превращаем в число
    except:
      area = 0 #Если не получается, возвращаем 0
  else:
    area = 0 #Или если нет разделителя, возвращаем 0
  return area

#Полуаем цену
def getCost(d):
  costStr = d[7] #Загружаем строку
  try:
    cost = float(costStr) #Пробуем превратить в число
  except:
    cost = 0 #Если не получается, возвращаем 0
  return cost

#Получаем комментарий
def getComment(d):
  commentStr = d[-1] #Возвращаем данные из последней колонки
  return commentStr

def words2Indexes(words, vocabulary, maxWordsCount):
  wordsIndexes = []
  for word in words:
    wordIndex = 0 #Тут будет индекс слова, изначально 0 - слово неопознано
    wordInVocabulary = word in vocabulary #Проверяем, есть ли слово в словаре
    if (wordInVocabulary):
      index = vocabulary[word] #Индекс = индексу слова в словаре
      if (index < maxWordsCount): #Если индекс ниже maxWordsCount - черты отсечения слов
        wordIndex = index #То записываем индекс
    wordsIndexes.append(wordIndex)
  return wordsIndexes

def changeSetTo01(trainSet, wordsCount):
  out = []
  for x in trainSet:
    out.append(changeXTo01(x, wordsCount)) #Добавляем в итоговый набор текущий вектор, преобразованный в bag of words
  return np.array(out)

def changeXTo01(trainVector, wordsCount):
  out = np.zeros(wordsCount)
  for x in trainVector:
    out[x] = 1 #В позицию нужного индекса ставим 1
  return out


from . import utils
from .dataset import Dataset
import seaborn as sns
sns.set_style('darkgrid') 

def загрузить_базу(база = ''):
    global ds
    ds = Dataset(база)
    ds.load_dataset()    

def показать_примеры():
    global ds
    ds.show_samples()
    
def создать_выборки():
    global ds
    ds.create_sampling()
    
def создать_сеть(слои):
    ds.create_model(слои)    
   
def обучить_модель(количество_эпох, количество_запусков=0):
    ds.train_model(количество_эпох, количество_запусков)
    
def тестировать_модель(data = None):
    ds.test_model(data)